#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include "xtensa/corebits.h"
#include "xtensa/config/core.h"
#include <sys/param.h>
#include "xtensa/specreg.h"

#include "esp_attr.h"
#include "esp_log.h"


typedef struct {
			uint8_t R;
			uint8_t G;
			uint8_t	B;
} color_t;

typedef struct {
			uint8_t A;
			uint8_t R;
			uint8_t G;
			uint8_t	B;
} colorA_t;

/**
  	@brief 
	The equation below is made from the original color mixing equation, to avoid using devision and fraction
	!!!This equation assumes that background is always not transparent!!!
	Accuracy of this is very good. 95% of time the 8bits of the color components are the same.
	Difference can maybe occur in the last LSB bit, but that does nearly nothing noticeable.
	
	This function should make the following equation (with order of execution)
	________III.________
				_____II._____
				___I.___     __IV.__
	((fg*fg.A)+(256-fg.A)*bg)/256
	The final devide by 256 is supplemented by a shifting to right by 8.
	Numerical testing shows this method is accurate enough.
	NO need to use slow floating point here.
  
	Where fg is the layers color component (R, G or B), bg is the background color
	Alpha is the alpha component of layer, bg has no alpha (non transparent)

	@param fg - Layer color with alpha channel in ARGB8888 format
	@param bg - Background color you want to mix onto RGB888 format
	
	@return result - RGB888 color
 **/
uint32_t cAblendARGB8888_WfWb(uint32_t fg, uint32_t bg);

/**
  	@brief 
	Same as cAlphaBlendARGB8888WO but with background color given as color_t
	here color_t must have .red, .green, .blue components, but no alpha

	@param fg - Layer color with alpha channel in ARGB8888 format
	@param bg - Background color you want to mix onto RGB888 format
	
	@return result - RGB888 color
 **/
uint32_t cAblendARGB8888_WfBb(uint32_t *fg, uint8_t bg_red, uint8_t bg_green, uint8_t bg_blue);

inline uint32_t cAblendARGB8888_BfBb(uint8_t fg_A, uint8_t fg_R, uint8_t fg_G, uint8_t fg_B, uint8_t bg_red, uint8_t bg_green, uint8_t bg_blue)
{
  //%0: result (output)

  //Register usage
  //A0: 256-fg.A
  //A1: fg.A
  //A2: - fg color word, and after formating, fg.B
  //A3: fg.G
  //A4: fg.R
  //A5: - bg color word, and then bg.B - and final BLUE
  //A6: bg.G - and final GREEN
  //A7: bg.R - and final RED
  
  
  //Opernad list
  //OUTPUTS:
  //%0 - result (output)
  //%1 - bg_red (outp)
  //%2 - bg_green (outp)
  //%3 - bg_blue (outp)
  //INPUTS
  //%4 - fg_A
  //%5 - fg_R
  //%6 - fg_G
  //%7 - fg_B
  //%8 - bg_red
  //%9 - bg_green
  //%10 - bg_blue
 	uint32_t result = 0;
	asm volatile (
	 //COMPONET RETRIEVAL
	 //256-fg.A	- I.
	 "MOVI a0, 255;"		
	 "SUB a0, a0, %4;"
	 
	 //HERE COMES THE BLENDING CODE
	 //RED
	 "MULL %1, a0, %8;" 	//1st multiplication -  II.
	 "MULL a1, %4, %5;"	//2nd multiplication, result in a1 - III.
	 "ADD %1, a1, %8;"		
	 //GREEN
	 "MULL %2, a0, %9;" 	//1st multiplication -  II.
	 "MULL a1, %4, %6;"	//2nd multiplication, result in a1 - III.
	 "ADD %2, a1, %9;"		
	 //BLUE
	 "MULL %3, a0, %10;" 	//1st multiplication -  II.
	 "MULL a1, %4, %7;"	//2nd multiplication, result in a1 - III.
	 "ADD %3, a1, %10;"		
	 
	 //IV. (Devision by 256 - RIGHT SHIFT BY 8).
	 //RSR instruction does not complete within 1 cycle, so it is advised to read RSR destination a couple cycles late
	 //this is why I put the shifting after the 3 RSR-s above
	 "SRLI %1, %8, 8;"
	 "SRLI %2, %9, 8;"
	 "SRLI %3, %10, 8;"
	 //At this point all RGB components are stored in registers A5-A7, with color value stored in bit 7-0.
	 //The next lines format and generate the result .
	 "SLLI %1, %8, 16;"
	 "SLLI %2, %9, 8;"
	 "ADD %1, %8, %9;"
	 "ADD %0, %8, %10;" 
	 
	 : "=r" (result), "=r" (bg_red), "=r" (bg_green), "=r" (bg_blue)
	 : "r" (fg_A), "r" (fg_R), "r" (fg_G), "r" (fg_B), "1" (bg_red), "2" (bg_green), "3" (bg_blue)
	 : "a0", "a1");
	 
	return result;
}
	
